#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Any, Dict, List, Sequence, Union

import ray
from ray import tune
from ray.tune.sample import Domain


class ComplexObject:
    """Class to store and retrieve large size objects and convert it to `ray.tune.Domain`.
    Objects will be stored with `ray.put` and retrieved with `ray.get`.
    """

    def __init__(self):
        self.values: List[Any] = []

    def __len__(self):
        return len(self.values)

    def append(self, value: Any):
        self.values.append(ray.put(value))

    def get_complex_object(self, idx):
        return ray.get(self.values[idx])

    def to_choice(self) -> Domain:
        """converts to ray.tune Domain"""
        indices = tuple(range(len(self.values)))
        return tune.choice(indices)


class Tuner:
    """
    Supports `ray.tune` methods and provide an easy way for tuning large size complex objects like Models.
    """

    def __init__(self):
        self._search_space: Dict[str, Domain] = {}
        self._complex_objects: Dict[str, ComplexObject] = {}

    def update_search_space(self, k: str, v: Union[Domain, ComplexObject]):
        """
        Update search space with value `ray.tune(...)` or `gradsflow.tuner.ComplexObject`
        Args:
            k: hyperparameter name
            v: hyperparameter value - `ray.tune(...)` or `gradsflow.tuner.ComplexObject` object.
        """
        if isinstance(v, Domain):
            self._search_space[k] = v
        elif isinstance(v, ComplexObject):
            assert isinstance(v, ComplexObject), f"Selected is_complex but object is of type {type(v)}"
            self._search_space[k] = v.to_choice()
            self._complex_objects[k] = v
        else:
            raise UserWarning(f"Tuner Search space doesn't support {type(v)}")
        assert isinstance(
            self._search_space[k], Domain
        ), "search space should only contain object of type `tune.Domain`"

    def suggest_complex(self, key: str, *values: Sequence) -> ComplexObject:
        """
        Use this method when you want to search models or any large object.
        It will also update search space with the provided key.
        Args:
            key: hyperparameter name
            *values: values for the hyperparameter

        Returns:
            `ComplexObject`
        """
        complex_object = ComplexObject()
        for i, v in enumerate(values):
            complex_object.append(v)

        object_choice = complex_object.to_choice()
        self._search_space[key] = object_choice
        self._complex_objects[key] = complex_object
        return complex_object

    def scalar(self, key: str, value):
        """This sets a scalar value and will not be used for tuning"""
        self._search_space[key] = value

    def choice(self, key: str, *values) -> Domain:
        """Tune for categorical values"""
        x = tune.choice(values)
        self._search_space[key] = x
        return x

    def loguniform(self, key: str, lower: float, upper: float, base: float = 10) -> Domain:
        x = tune.loguniform(lower, upper, base)
        self._search_space[key] = x
        return x

    def union(self, tuner: "Tuner") -> "Tuner":
        """Inplace Merge of two Tuners"""
        self._search_space.update(tuner._search_space)
        self._complex_objects.update(tuner._complex_objects)
        return self

    @staticmethod
    def merge(*tuners: "Tuner") -> "Tuner":
        final_tuner = Tuner()
        for t in tuners:
            final_tuner.union(t)
        return final_tuner

    @property
    def value(self) -> Dict[str, Domain]:
        return self._search_space

    def get_complex_object(self, key: str, idx: int):
        """Get registered complex object value from key at given index"""
        return self._complex_objects[key].get_complex_object(idx)

    def get(self, key: str) -> Union[Domain, ComplexObject]:
        if key in self._complex_objects:
            return self._complex_objects[key]
        if key in self._search_space:
            return self._search_space[key]
        raise KeyError(f"key={key} is not available in tuner! Available={tuple(self._search_space.keys())}")



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os

os.environ["GF_CI"] = "true"

import torch.nn
from ray import tune
from timm import create_model

from gradsflow.core.data import AutoDataset
from gradsflow.data.image import get_fake_data
from gradsflow.models.constants import LEARNER
from gradsflow.tuner import AutoModelV2 as AutoModel
from gradsflow.tuner import Tuner

image_size = (64, 64)
train_data = get_fake_data(image_size, num_classes=2)
val_data = get_fake_data(image_size, num_classes=2)

num_classes = train_data.dataset.num_classes
autodataset = AutoDataset(train_data.dataloader, val_data.dataloader, num_classes=num_classes)


def test_hp_tune():
    tuner = Tuner()
    cnn = create_model("resnet18", pretrained=False, num_classes=num_classes)

    model = AutoModel(cnn, optimization_metric="val_loss")
    model.compile(
        loss="crossentropyloss", optimizer=tune.choice(("adam", "sgd")), learning_rate=tune.loguniform(1e-5, 1e-3)
    )

    model.hp_tune(
        tuner,
        autodataset,
        n_trials=1,
        epochs=1,
        cpu=0.05,
        gpu=0,
        trainer_config={"steps_per_epoch": 2},
    )


def test_get_learner():
    tuner = Tuner()
    cnn = create_model("resnet18", pretrained=False, num_classes=num_classes)
    complex_cnn = tuner.suggest_complex("learner", cnn)
    automodel = AutoModel(complex_cnn, optimization_metric="val_loss")
    hparams = {LEARNER: 0}
    model = automodel._get_learner(hparams, tuner)
    assert isinstance(model, torch.nn.Module)


def test_compile():
    tuner = Tuner()
    cnn = create_model("resnet18", pretrained=False, num_classes=num_classes)
    complex_cnn = tuner.suggest_complex("learner", cnn)

    model = AutoModel(complex_cnn, optimization_metric="val_loss")
    model.compile(
        loss="crossentropyloss", optimizer=tune.choice(("adam", "sgd")), learning_rate=tune.loguniform(1e-5, 1e-3)
    )



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest
from rich.table import Table

from gradsflow.models.tracker import Tracker

tracker = Tracker()


def test_reset():
    tracker.max_epochs = 5
    tracker.reset()
    assert tracker.max_epochs == 0


def test_mode():
    tracker.mode("train")
    tracker.mode("val")
    with pytest.raises(NotImplementedError):
        tracker.mode("test")


def test_track():
    tracker.track("val", 0.9, render=True)
    tracker.track("score", 0.5, render=False)


def test_create_table():
    tracker.track_loss(0.1, "train")
    tracker.track_loss(0.2, "val")
    table = tracker.create_table()
    assert isinstance(table, Table)



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from abc import ABC
from typing import Optional, Union

import pytorch_lightning as pl
import torch
from ray import tune
from torch.utils.data import DataLoader

from gradsflow.core.backend import AutoBackend
from gradsflow.core.base import BaseAutoModel
from gradsflow.core.data import AutoDataset
from gradsflow.utility.common import module_to_cls_index

logger = logging.getLogger("core.model")


class AutoModel(BaseAutoModel, ABC):
    """
    Base model that defines hyperparameter search methods and initializes `Ray`.
    All other tasks are implementation of `AutoModel`.

    Args:
        datamodule flash.DataModule: DataModule from Flash or PyTorch Lightning
        max_epochs [int]: Maximum number of epochs for which model will train
        max_steps Optional[int]: Maximum number of steps for each epoch. Defaults None.
        optimization_metric str: Value on which hyperparameter search will run.
        By default, it is `val_accuracy`.
        n_trials int: Number of trials for HPO
        suggested_conf Dict: Any extra suggested configuration
        timeout int: HPO will stop after timeout
        prune bool: Whether to stop unpromising training.
        backend Optional[str]: Training backend - PL / torch / fastai. Default is PL
    """

    _OPTIMIZER_INDEX = module_to_cls_index(torch.optim, True)
    _DEFAULT_OPTIMIZERS = ["adam", "sgd"]
    _DEFAULT_LR = (1e-5, 1e-2)

    def __init__(
        self,
        datamodule: Optional[pl.LightningDataModule] = None,
        train_dataloader: Optional[DataLoader] = None,
        val_dataloader: Optional[DataLoader] = None,
        num_classes: Optional[int] = None,
        max_epochs: int = 10,
        max_steps: Optional[int] = None,
        optimization_metric: Optional[str] = None,
        n_trials: int = 20,
        suggested_conf: Optional[dict] = None,
        timeout: int = 600,
        prune: bool = True,
        backend: Optional[str] = None,
    ):

        self.analysis = None
        self.prune = prune
        self.n_trials = n_trials
        self.model: Union[torch.nn.Module, pl.LightningModule, None] = None
        self.max_epochs = max_epochs
        self.max_steps = max_steps
        self.timeout = timeout
        self.optimization_metric = optimization_metric or "val_accuracy"
        self.suggested_conf = suggested_conf or {}

        self.auto_dataset = AutoDataset(
            train_dataloader=train_dataloader,
            val_dataloader=val_dataloader,
            datamodule=datamodule,
            num_classes=num_classes,
        )

        self.datamodule = self.auto_dataset.datamodule

        self.backend = AutoBackend(
            self.auto_dataset,
            model_builder=self.build_model,
            optimization_metric=optimization_metric,
            max_epochs=max_epochs,
            max_steps=max_steps,
            backend=backend,
        )

        self.suggested_optimizers = self.suggested_conf.get("optimizer", self._DEFAULT_OPTIMIZERS)
        default_lr = self._DEFAULT_LR
        self.suggested_lr = self.suggested_conf.get("lr") or self.suggested_conf.get("learning_rate") or default_lr

    def hp_tune(
        self,
        name: Optional[str] = None,
        ray_config: Optional[dict] = None,
        trainer_config: Optional[dict] = None,
        mode: Optional[str] = None,
        gpu: Optional[float] = 0,
        cpu: Optional[float] = None,
        resume: bool = False,
    ):
        """
        Search Hyperparameter and builds model with the best params

        ```python
            automodel = AutoClassifier(data)  # implements `AutoModel`
            automodel.hp_tune(name="gflow-example", gpu=1)
        ```

        Args:
            name Optional[str]: name of the experiment.
            ray_config dict: configuration passed to `ray.tune.run(...)`
            trainer_config dict: configuration passed to `pl.trainer.fit(...)`
            mode Optional[str]: Whether to maximize or minimize the `optimization_metric`.
            Values are `max` or `min`
            gpu Optional[float]: Amount of GPU resource per trial.
            cpu float: CPU cores per trial
            resume bool: Whether to resume the training or not.
        """

        trainer_config = trainer_config or {}
        ray_config = ray_config or {}

        search_space = self._create_search_space()
        trainable = self.backend.optimization_objective

        resources_per_trial = {}
        if gpu:
            resources_per_trial["gpu"] = gpu
        if cpu:
            resources_per_trial["cpu"] = cpu

        mode = mode or "max"
        timeout_stopper = tune.stopper.TimeoutStopper(self.timeout)
        logger.info(
            "tuning _search_space with metric = {} and model = {}".format(
                self.optimization_metric,
                mode,
            )
        )
        analysis = tune.run(
            tune.with_parameters(trainable, trainer_config=trainer_config),
            name=name,
            num_samples=self.n_trials,
            metric=self.optimization_metric,
            mode=mode,
            config=search_space,
            resources_per_trial=resources_per_trial,
            resume=resume,
            stop=timeout_stopper,
            checkpoint_at_end=True,
            **ray_config,
        )
        self.analysis = analysis
        self.model = self._get_best_model(analysis)

        logger.info(f"🎉 Best hyperparameters found were: {analysis.best_config}")
        return analysis

    def _get_best_model(self, analysis):

        best_model = self.build_model(self.analysis.best_config)
        best_ckpt_path = analysis.best_checkpoint
        best_ckpt_path = best_ckpt_path + "/filename"
        best_model = best_model.load_from_checkpoint(best_ckpt_path)
        return best_model



# Copyright (c) 2021 GradsFlow. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

os.environ["GF_CI"] = "true"

import warnings
from pathlib import Path

import pytest
import torch

from gradsflow.data.image import image_dataset_from_directory
from gradsflow.models.model import Model
from gradsflow.tasks import AutoImageClassifier

warnings.filterwarnings("ignore")

data_dir = Path.cwd() / "data"

train_data = image_dataset_from_directory(f"{data_dir}/hymenoptera_data/train/", transform=True)
train_dl = train_data.dataloader

val_data = image_dataset_from_directory(f"{data_dir}/hymenoptera_data/val/", transform=True)
val_dl = val_data.dataloader


def test_forward():
    model = AutoImageClassifier(
        train_dataloader=train_dl,
        val_dataloader=val_dl,
        num_classes=2,
    )

    with pytest.raises(UserWarning):
        model.forward(torch.rand(1, 3, 8, 8))


def test_build_model():
    automodel = AutoImageClassifier(
        train_dataloader=train_dl,
        val_dataloader=val_dl,
        num_classes=2,
        max_epochs=1,
        max_steps=5,
        timeout=10,
        suggested_backbones="ssl_resnet18",
        n_trials=1,
    )
    kwargs = {"backbone": "ssl_resnet18", "optimizer": "adam", "lr": 1e-1}
    automodel.model = automodel.build_model(kwargs)
    assert isinstance(automodel.model, Model)


def test_hp_tune():
    model = AutoImageClassifier(
        train_dataloader=train_dl,
        val_dataloader=val_dl,
        num_classes=2,
        max_epochs=1,
        max_steps=5,
        timeout=10,
        suggested_backbones="ssl_resnet18",
        optimization_metric="train_loss",
        n_trials=1,
    )
    model.hp_tune(name="pytest-experiment", mode="min", gpu=0)



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import inspect
import os
from enum import Enum
from typing import Any, Dict, Optional, Union

import ray
from loguru import logger
from ray import tune
from ray.tune.sample import Domain
from torch import nn

from gradsflow.core.data import AutoDataset
from gradsflow.models import Model
from gradsflow.models.constants import LEARNER
from gradsflow.models.model import METRICS_TYPE
from gradsflow.tuner.tuner import ComplexObject, Tuner


class State(Enum):
    TUNER = "tuner"


class AutoModelV2:
    """Search Hyperparameter for your `Model`

    Examples:
        ```python
        tuner = Tuner()
        cnns = tuner.suggest_complex("learner", cnn1, cnn2)
        optimizers = tuner.choice("optimizer", "adam", "sgd")
        loss = "crossentropyloss"
        model = AutoModelV2(cnns)
        model.hp_tune(tuner, autodataset, max_epochs=10)
        ```

    Args:
        learner: tuner.
        optimization_metric: metric on which to optimize model on
        mode: max or min for optimization_metric

    """

    TEST = os.environ.get("GF_CI", "false").lower() == "true"

    def __init__(
        self,
        learner: Union[ComplexObject, nn.Module],
        optimization_metric: Optional[str] = None,
        mode: Optional[str] = None,
    ):
        self.tuner: Tuner = Tuner()
        self.learner = self._register_model(LEARNER, learner)
        self.optimization_metric = optimization_metric or "train_loss"
        self.mode = mode or "min"
        self.analysis = None
        self._non_domain_config = {"compile": {}}

    def _register_model(self, object_name: str, variable: Any):
        if isinstance(variable, (Domain, ComplexObject)):
            self.tuner.update_search_space(object_name, variable)
            return State.TUNER
        return ray.put(variable)

    def compile(
        self,
        loss: Union[str, Domain] = None,
        optimizer: Union[str, Domain] = None,
        learning_rate: Union[float, Domain] = 3e-4,
        optimizer_config: Union[Dict[str, Domain], None] = None,
        metrics: METRICS_TYPE = None,
    ) -> None:
        compile_config = locals()
        for k, v in compile_config.items():
            logger.debug(f"compile: {k}, {v}")
            if k in ("self", "compile"):
                continue
            if isinstance(v, Domain):
                self.tuner.update_search_space(k, v)
            else:
                self._non_domain_config["compile"][k] = v

    def _get_learner(self, hparams, tuner):
        """Fetch learner from tuner or self.learner"""
        if self.learner == State.TUNER:
            idx = hparams.get(LEARNER)
            learner = tuner.get(LEARNER).get_complex_object(idx)
        else:
            learner = ray.get(self.learner)

        return learner

    def build_model(self, hparams, tuner) -> Model:
        """build and compile Model"""
        learner = self._get_learner(hparams, tuner)
        model = Model(learner)

        compile_args = set(inspect.getfullargspec(model.compile).args)
        compile_args.remove("self")
        domain_compile_args = {k: v for k, v in hparams.items() if k in compile_args}
        logger.debug(self._non_domain_config)
        logger.debug(domain_compile_args)

        model.compile(
            **self._non_domain_config["compile"],
            **domain_compile_args,
        )

        return model

    def trainable(
        self, search_space: Dict[str, Domain], autodataset: AutoDataset, epochs: int, tuner: Tuner, fit_config: dict
    ):
        model: Model = self.build_model(search_space, tuner)

        model.fit(
            autodataset,
            max_epochs=epochs,
            callbacks=["tune_checkpoint", "tune_report"],
            show_progress=False,
            **fit_config,
        )

    def hp_tune(
        self,
        tuner: Tuner,
        autodataset,
        epochs: int = 1,
        n_trials=10,
        gpu=None,
        cpu=None,
        time: int = None,
        resume=False,
        trainer_config: Optional[dict] = None,
        ray_config: Optional[dict] = None,
    ):
        trainer_config = trainer_config or {}
        resources_per_trial = {}
        ray_config = ray_config or {}
        if gpu:
            resources_per_trial["gpu"] = gpu
        if cpu:
            resources_per_trial["cpu"] = cpu
        self.tuner.union(tuner)
        search_space = self.tuner.value
        analysis = tune.run(
            tune.with_parameters(
                self.trainable, autodataset=autodataset, epochs=epochs, tuner=tuner, fit_config=trainer_config
            ),
            num_samples=n_trials,
            metric=self.optimization_metric,
            mode=self.mode,
            config=search_space,
            resources_per_trial=resources_per_trial,
            resume=resume,
            time_budget_s=time,
            checkpoint_at_end=True,
            **ray_config,
        )
        self.analysis = analysis
        logger.info(f"🎉 Best HyperParameters found: {analysis.best_config}")



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Union

import torch
from accelerate import Accelerator
from torch import nn
from torchmetrics import Metric, MetricCollection

from gradsflow.core.data import AutoDataset
from gradsflow.models.tracker import Tracker
from gradsflow.models.utils import get_device, losses
from gradsflow.models.utils import metrics as metrics_classes
from gradsflow.utility.common import module_to_cls_index


@dataclass(init=False)
class Base:
    learner: Union[nn.Module, Any]
    optimizer: torch.optim.Optimizer = None
    loss: Callable = None
    _compiled: bool = False


class BaseModel(Base):
    """Base Class of Model API"""

    TEST = os.environ.get("GF_CI", "false").lower() == "true"
    _OPTIMIZER_INDEX = module_to_cls_index(torch.optim, True)

    def __init__(
        self,
        learner: Union[nn.Module, Any],
        device: Optional[str] = None,
        use_accelerate: bool = True,
        accelerator_config: dict = None,
    ):
        self.tracker = Tracker()
        self.accelerator = None
        self.device = None
        self._set_accelerator(device, use_accelerate, accelerator_config)
        self.learner = self.prepare_model(learner)
        self.metrics: MetricCollection = MetricCollection([])

    def _set_accelerator(self, device: Optional[str], use_accelerate: bool, accelerator_config: dict):
        if use_accelerate:
            self.accelerator = Accelerator(cpu=(device == "cpu"), **accelerator_config)
            self.device = self.accelerator.device
        else:
            self.device = device or get_device()

    def prepare_model(self, learner: Union[nn.Module, List[nn.Module]]):
        """Inplace ops for preparing model via HF Accelerator. Automatically sends to device."""
        if not self.accelerator:
            learner = learner.to(self.device)
            return learner
        if isinstance(learner, (list, tuple)):
            self.learner = list(map(self.accelerator.prepare_model, learner))
        elif isinstance(learner, nn.Module):
            self.learner = self.accelerator.prepare_model(learner)
        else:
            raise NotImplementedError(
                f"prepare_model is not implemented for model of type {type(learner)}! Please implement prepare_model "
                f"or raise an issue."
            )

        return self.learner

    def prepare_optimizer(self, optimizer) -> torch.optim.Optimizer:
        if not self.accelerator:
            return optimizer
        return self.accelerator.prepare_optimizer(optimizer)

    def prepare_data(self, autodataset: AutoDataset = None) -> AutoDataset:
        if not self.accelerator:
            return autodataset
        autodataset.train_dataloader = self.accelerator.prepare_data_loader(autodataset.train_dataloader)
        if autodataset.val_dataloader:
            autodataset.val_dataloader = self.accelerator.prepare_data_loader(autodataset.val_dataloader)
        autodataset.sent_to_device = True
        return autodataset

    def add_metrics(self, *metrics: Union[str, Metric]) -> None:
        for m in metrics:
            if isinstance(m, str):
                m_cls = metrics_classes.get(m)
                assert (
                    m_cls is not None
                ), f"metrics {m} is not available! Available metrics are {tuple(metrics_classes.keys())}"
                m_obj = m_cls()
            elif isinstance(m, Metric):
                m_obj = m
            else:
                raise NotImplementedError(f"metrics not implemented for {m}! Please see `torchmetrics`.")
            self.metrics.add_metrics(m_obj)

    def _get_loss(self, loss: Union[str, Callable]) -> Optional[Callable]:
        loss_fn = None
        if isinstance(loss, str):
            loss_fn = losses.get(loss)
            assert loss_fn is not None, f"loss {loss} is not available! Available losses are {tuple(losses.keys())}"
        elif callable(loss):
            loss_fn = loss

        return loss_fn

    def _get_optimizer(self, optimizer: Union[str, torch.optim.Optimizer]) -> Callable:
        if isinstance(optimizer, str):
            optimizer_fn = self._OPTIMIZER_INDEX.get(optimizer)
            assert (
                optimizer_fn
            ), f"optimizer {optimizer} is not available! Available optimizers are {tuple(self._OPTIMIZER_INDEX.keys())}"

        elif callable(optimizer):
            assert optimizer in tuple(self._OPTIMIZER_INDEX.values()), f"Unknown Optimizer {type(optimizer)}"
            optimizer_fn = optimizer
        else:
            raise NotImplementedError(f"Unknown optimizer {optimizer}")
        return optimizer_fn

    def assert_compiled(self):
        if not self._compiled:
            raise UserWarning("Model not compiled yet! Please call `model.compile(...)` first.")

    def forward(self, x):
        return self.learner(x)

    def __call__(self, x):
        return self.forward(x)

    @torch.no_grad()
    def predict(self, x):
        return self.learner(x)

    def load_from_checkpoint(self, checkpoint):
        self.learner = torch.load(checkpoint)

    def backward(self, loss: torch.Tensor):
        if not self.accelerator:
            loss.backward()
        else:
            self.accelerator.backward(loss)

    def eval(self):
        """Set learner to eval mode for validation"""
        self.learner.requires_grad_(False)
        self.learner.eval()

    def train(self):
        """Set learner to training mode"""
        self.learner.requires_grad_(True)
        self.learner.train()



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Dict, List, Union


class DataMixin:
    @staticmethod
    def fetch_inputs(data: Union[List, Dict]):
        return data[0]

    @staticmethod
    def fetch_target(data: Union[List, Dict]):
        return data[1]

    @staticmethod
    def send_to_device(batch: Union[List, Dict], device):
        """Send data to be device"""
        if isinstance(batch, (list, tuple)):
            return list(map(lambda x: x.to(device), batch))
        if isinstance(batch, dict):
            return {k: v.to(device) for k, v in batch.items()}
        raise NotImplementedError(
            f"send_to_device is not implemented for data of type {type(batch)}! Please raise an issue/pr"
        )



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest
import torch
from torch.utils.data import DataLoader, TensorDataset

from gradsflow.core.data import AutoDataset

dataset = TensorDataset(torch.randn(8, 1, 32, 32))
dataloader = DataLoader(dataset)


def test_auto_dataset():
    with pytest.raises(UserWarning):
        AutoDataset()


def test_sent_to_device():
    data = AutoDataset(dataloader)
    assert data.sent_to_device is None
    data.sent_to_device = True
    assert data.sent_to_device



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest
import timm
import torch

from gradsflow.core.data import AutoDataset
from gradsflow.data.image import get_fake_data
from gradsflow.models.model import Model
from gradsflow.models.tracker import Tracker

image_size = (64, 64)
train_data = get_fake_data(image_size)
val_data = get_fake_data(image_size)

num_classes = train_data.dataset.num_classes
autodataset = AutoDataset(train_data.dataloader, val_data.dataloader, num_classes=num_classes)

cnn = timm.create_model("ssl_resnet18", pretrained=False, num_classes=num_classes).eval()
model = Model(cnn)
model.compile("crossentropyloss", "adam")
model.TEST = True


def test_predict():
    x = torch.randn(1, 3, 64, 64)
    r1 = model.forward(x)
    r2 = model(x)
    r3 = model.predict(x)
    assert torch.all(torch.isclose(r1, r2))
    assert torch.all(torch.isclose(r2, r3))
    assert isinstance(model.predict(torch.randn(1, 3, 64, 64)), torch.Tensor)


def test_fit():
    model.TEST = True
    tracker = model.fit(autodataset, max_epochs=1, steps_per_epoch=1, show_progress=True)
    assert isinstance(tracker, Tracker)


def test_compile():
    model1 = Model(cnn)

    def compute_accuracy(*_, **__):
        return 1

    with pytest.raises(NotImplementedError):
        model1.compile("crossentropyloss", "adam", metrics=compute_accuracy)

    with pytest.raises(AssertionError):
        model1.compile("crossentropyloss", "adam", metrics="random_val")

    model1.compile("crossentropyloss", "adam", metrics="accuracy")

    model2 = Model(cnn)
    model2.compile("crossentropyloss", torch.optim.Adam)
    model2.compile(torch.nn.CrossEntropyLoss, torch.optim.Adam, learning_rate=0.01)
    assert model2.optimizer.param_groups[0]["lr"] == 0.01


def test_set_accelerator():
    model2 = Model(cnn, accelerator_config={"fp16": True})
    model2.compile()
    assert model2.accelerator



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import typing
from typing import Union

from gradsflow.callbacks.callbacks import Callback
from gradsflow.callbacks.progress import ProgressCallback
from gradsflow.callbacks.raytune import TorchTuneCheckpointCallback, TorchTuneReport

if typing.TYPE_CHECKING:
    from gradsflow.models.model import Model


class CallbackRunner(Callback):
    _AVAILABLE_CALLBACKS = {
        "tune_checkpoint": TorchTuneCheckpointCallback,
        "tune_report": TorchTuneReport,
        "progress": ProgressCallback,
    }

    def __init__(self, model: "Model", *callbacks: Union[str, Callback]):
        super().__init__(model)
        self.callbacks = []
        for callback in callbacks:
            if isinstance(callback, str):
                callback_fn = self._AVAILABLE_CALLBACKS[callback](model)
                self.callbacks.append(callback_fn)
            elif isinstance(callback, Callback):
                self.callbacks.append(callback)
            else:
                raise NotImplementedError

    def available_callbacks(self):
        return list(self._AVAILABLE_CALLBACKS.keys())

    def on_train_epoch_end(self):
        for callback in self.callbacks:
            callback.on_train_epoch_end()

    def on_train_epoch_start(self):
        for callback in self.callbacks:
            callback.on_train_epoch_start()

    def on_fit_start(self):
        for callback in self.callbacks:
            callback.on_fit_start()

    def on_fit_end(
        self,
    ):
        for callback in self.callbacks:
            callback.on_fit_end()

    def on_val_epoch_start(
        self,
    ):
        for callback in self.callbacks:
            callback.on_val_epoch_start()

    def on_val_epoch_end(self):
        for callback in self.callbacks:
            callback.on_val_epoch_end()

    def on_train_step_start(self):
        for callback in self.callbacks:
            callback.on_train_step_start()

    def on_train_step_end(self):
        for callback in self.callbacks:
            callback.on_train_step_end()

    def on_val_step_start(self):
        for callback in self.callbacks:
            callback.on_val_step_start()

    def on_val_step_end(self):
        for callback in self.callbacks:
            callback.on_val_step_end()

    def on_epoch_start(self):
        for callback in self.callbacks:
            callback.on_epoch_start()

    def on_epoch_end(self):
        for callback in self.callbacks:
            callback.on_epoch_end()

    def on_forward_start(self):
        for callback in self.callbacks:
            callback.on_forward_start()

    def on_forward_end(self):
        for callback in self.callbacks:
            callback.on_forward_end()

    def clean(self):
        for callback in self.callbacks:
            callback.clean()



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Dict, List, Optional

from loguru import logger
from rich.table import Table

from gradsflow.core.base import BaseTracker, TrackingValues
from gradsflow.core.data import AutoDataset
from gradsflow.models.utils import to_item


class Tracker(BaseTracker):
    """
    Tracks loss, accuracy and model weights during model.fit()
    """

    def __init__(self):
        self.learner = None
        self.autodataset: Optional[AutoDataset] = None
        self.train.metrics = {}
        self.val.metrics = {}
        self.logs: List[Dict] = []
        self.non_render_logs: List[Dict] = []

    def mode(self, mode) -> TrackingValues:
        if mode == "train":
            return self.train
        if mode == "val":
            return self.val

        raise NotImplementedError(f"mode {mode} is not implemented!")

    def track(self, key, value, render=False):
        if render:
            logger.warning("render is deprecated!")
        epoch = self.current_epoch
        step = self.current_step
        data = {"current_epoch": epoch, "current_step": step, key: to_item(value)}
        self.logs.append(data)

    def track_loss(self, loss: float, mode: str):
        """Update `TrackingValues` loss. mode can be train or val"""
        value_tracker = self.mode(mode)
        value_tracker.loss = loss
        key = mode + "/" + "loss"
        self.track(key, loss)

    def track_metrics(self, metric: Dict[str, float], mode: str, render: bool = False):
        """Update `TrackingValues` metrics. mode can be train or val and will update logs if render is True"""
        value_tracker = self.mode(mode)
        value_tracker.metrics = metric

        for k, v in metric.items():
            k = mode + "/" + k
            self.track(k, v, render=render)

    def get_metrics(self, mode: str):
        value_tracker = self.mode(mode)
        return value_tracker.metrics

    def get_loss(self, mode: str):
        value_tracker = self.mode(mode)
        return value_tracker.loss

    def create_table(self) -> Table:
        headings = ["current_epoch", "loss", "train/loss"]
        row = [self.current_epoch, to_item(self.train.step_loss), to_item(self.train.loss)]
        if self.val.loss:
            headings.append("val/loss")

        for metric_name, _ in self.train.metrics.items():
            headings.append("train/" + metric_name)

        for metric_name, _ in self.val.metrics.items():
            headings.append("val/" + metric_name)

        if self.val.loss:
            row.append(to_item(self.val.loss))

        for _, value in self.train.metrics.items():
            row.append(to_item(value))

        for _, value in self.val.metrics.items():
            row.append(to_item(value))

        row = list(map(lambda x: f"{x: .3f}" if isinstance(x, float) else str(x), row))
        table = Table(*headings, expand=False)
        table.add_row(*row)
        return table

    def reset(self):
        self.max_epochs = 0
        self.current_epoch = 0
        self.current_step = 0
        self.steps_per_epoch = None
        self.train = TrackingValues()
        self.val = TrackingValues()
        self.logs = []
        self.non_render_logs = []



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import numpy as np
import torch

from gradsflow.models.utils import get_device, to_item


def test_to_item():
    x = torch.rand(1, 1, requires_grad=True)
    assert isinstance(to_item(x), np.ndarray)


def test_get_device():
    assert get_device() in ("cpu", "cuda")



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

from flash.image import ImageClassificationData

from gradsflow import AutoImageClassifier
from gradsflow.data.image import image_dataset_from_directory

# 1. Create the DataModule
data_dir = os.getcwd() + "/data"
# download_data("https://pl-flash-data.s3.amazonaws.com/hymenoptera_data.zip", data_dir)

datamodule = ImageClassificationData.from_folders(
    train_folder=f"{data_dir}/hymenoptera_data/train/",
    val_folder=f"{data_dir}/hymenoptera_data/val/",
)
image_size = (64, 64)
train_data = image_dataset_from_directory(f"{data_dir}/hymenoptera_data/train/", image_size=image_size, transform=True)

val_data = image_dataset_from_directory(f"{data_dir}/hymenoptera_data/val/", image_size=image_size, transform=True)

num_classes = len(train_data.dataset.classes)

model = AutoImageClassifier(
    train_dataloader=train_data.dataloader,
    val_dataloader=val_data.dataloader,
    num_classes=num_classes,
    max_epochs=1,
    optimization_metric="val_loss",
    max_steps=5,
    n_trials=1,
)
print("AutoImageClassifier initialised!")

model.hp_tune()



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
from typing import Optional

import torch
from ray import tune
from ray.tune.integration.pytorch_lightning import TuneReportCheckpointCallback

from .callbacks import Callback

_METRICS = {
    "val_accuracy": "val_accuracy",
    "train_accuracy": "train_accuracy",
}


def report_checkpoint_callback(metrics: Optional[dict] = None, filename: Optional[str] = None):
    metrics = metrics or _METRICS
    filename = filename or "filename"
    callback = TuneReportCheckpointCallback(metrics=metrics, filename=filename, on="validation_end")

    return callback


class TorchTuneCheckpointCallback(Callback):
    def on_epoch_end(self):
        epoch = self.model.tracker.current_epoch
        model = self.model.learner

        with tune.checkpoint_dir(epoch) as checkpoint_dir:
            path = os.path.join(checkpoint_dir, "filename")
            torch.save((model.state_dict()), path)


class TorchTuneReport(Callback):
    def on_epoch_end(self):
        val_loss = self.model.tracker.val.loss
        train_loss = self.model.tracker.train.loss
        val_tracker = self.model.tracker.train.metrics
        train_tracker = self.model.tracker.val.metrics

        train_metrics = {"train_" + k.lower(): v for k, v in train_tracker.items()}
        val_metrics = {"val_" + k.lower(): v for k, v in val_tracker.items()}

        tune.report(val_loss=val_loss, train_loss=train_loss, **train_metrics, **val_metrics)



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import torch
from flash.text.seq2seq import SummarizationTask

from gradsflow.core.autoclassifier import AutoClassifier


# noinspection PyTypeChecker
class AutoSummarization(AutoClassifier):
    """
    Automatically finds Text Summarization Model

    Args:
        datamodule Optional[DataModule]: PL Lightning DataModule with `num_classes` property.
        train_dataloader Optional[DataLoader]: torch dataloader
        val_dataloader Optional[DataLoader]: torch dataloader
        num_classes Optional[int]: number of classes
        max_epochs [int]: default=10.
        n_trials [int]: default=100.
        optimization_metric [Optional[str]]: defaults None
        suggested_backbones Union[List, str, None]: defaults None
        suggested_conf [Optional[dict] = None]: This sets Trial suggestions for optimizer,
            learning rate, and all the hyperparameters.
        timeout [int]: Hyperparameter search will stop after timeout.

    Examples:
        ```python
            from gradsflow import AutoSummarization

            from flash.core.data.utils import download_data
            from flash.text import SummarizationData, SummarizationTask

            # 1. Download the data
            download_data("https://pl-flash-data.s3.amazonaws.com/xsum.zip", "data/")
            # 2. Load the data
            datamodule = SummarizationData.from_csv(
                "input",
                "target",
                train_file="data/xsum/train.csv",
                val_file="data/xsum/valid.csv",
                test_file="data/xsum/test.csv",
            )

            model = AutoSummarization(datamodule,
                                    max_epochs=10,
                                    optimization_metric="val_accuracy",
                                    timeout=300)
            model.hp_tune()
        ```
    """

    _DEFAULT_BACKBONES = [
        "sshleifer/distilbart-cnn-12-6",
        "sshleifer/distilbart-xsum-12-3",
    ]

    def build_model(self, config: dict) -> torch.nn.Module:
        """Build SummarizationModel from `ray.tune` hyperparameter configs
        or via _search_space dictionary arguments

        Arguments:
            backbone [str]: Image classification backbone name -
            sshleifer/distilbart-cnn-12-6, sshleifer/distilbart-xsum-12-3,...
            (Check Lightning-Flash for full model list)

            optimizer [str]: PyTorch Optimizers. Check `AutoImageClassification._OPTIMIZER_INDEX`
            learning_rate [float]: Learning rate for the model.
        """
        backbone = config["backbone"]
        optimizer = config["optimizer"]
        learning_rate = config["lr"]

        return SummarizationTask(
            backbone=backbone,
            optimizer=self._OPTIMIZER_INDEX[optimizer],
            learning_rate=learning_rate,
        )



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import torchvision
from timm import create_model
from torch.utils.data import DataLoader
from torchvision import transforms as T

from gradsflow import AutoDataset, Model
from gradsflow.data.common import random_split_dataset

# Replace dataloaders with your custom dataset and you are all set to train your model
image_size = (64, 64)
batch_size = 4

to_rgb = lambda x: x.convert("RGB")

augs = T.Compose([to_rgb, T.AutoAugment(), T.Resize(image_size), T.ToTensor()])
data = torchvision.datasets.Caltech101("~/", download=True, transform=augs)
train_data, val_data = random_split_dataset(data, 0.99)
train_dl = DataLoader(train_data, batch_size=batch_size)
val_dl = DataLoader(val_data, batch_size=batch_size)
num_classes = len(data.categories)

if __name__ == "__main__":
    autodataset = AutoDataset(train_dl, val_dl, num_classes=num_classes)
    cnn = create_model("resnet18", pretrained=False, num_classes=num_classes)

    model = Model(cnn)
    model.compile("crossentropyloss", "adam", metrics=["accuracy"])
    model.fit(autodataset, max_epochs=10, steps_per_epoch=50)



import os

os.environ["GF_CI"] = "true"

import warnings
from pathlib import Path

import ray
from flash.image import ImageClassificationData

from gradsflow.data.image import image_dataset_from_directory
from gradsflow.models.model import Model
from gradsflow.tasks import autotask

warnings.filterwarnings("ignore")

ray.init(local_mode=True, ignore_reinit_error=True)

data_dir = Path.cwd()
datamodule = ImageClassificationData.from_folders(
    train_folder=f"{data_dir}/data/hymenoptera_data/train/",
    val_folder=f"{data_dir}/data/hymenoptera_data/val/",
)
data_dir = Path.cwd() / "data"

train_data = image_dataset_from_directory(f"{data_dir}/hymenoptera_data/train/", transform=True)
train_dl = train_data.dataloader

val_data = image_dataset_from_directory(f"{data_dir}/hymenoptera_data/val/", transform=True)
val_dl = val_data.dataloader


def test_build_model():
    model = autotask(
        task="image-classification",
        train_dataloader=train_dl,
        val_dataloader=val_dl,
        num_classes=2,
        timeout=5,
        suggested_backbones="ssl_resnet18",
        n_trials=1,
    )
    kwargs = {"backbone": "ssl_resnet18", "optimizer": "adam", "lr": 1e-1}
    model.model = model.build_model(kwargs)
    assert isinstance(model.model, Model)


def test_hp_tune():
    model = autotask(
        task="image-classification",
        train_dataloader=train_dl,
        val_dataloader=val_dl,
        num_classes=2,
        max_epochs=1,
        max_steps=2,
        timeout=10,
        suggested_backbones="ssl_resnet18",
        optimization_metric="val_loss",
        n_trials=1,
    )
    model.hp_tune(name="pytest-experiment", mode="max", gpu=0)



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import dataclasses
import logging
from typing import Callable, Optional

import pytorch_lightning as pl
from torch.utils.data import DataLoader, Dataset

from gradsflow.data.ray_dataset import RayDataset

logger = logging.getLogger("core.data")


@dataclasses.dataclass(init=False)
class Data:
    dataloader: DataLoader
    dataset: [RayDataset, Dataset]


class AutoDataset:
    def __init__(
        self,
        train_dataloader: Optional[DataLoader] = None,
        val_dataloader: Optional[DataLoader] = None,
        datamodule: Optional[pl.LightningDataModule] = None,
        num_classes: Optional[int] = None,
    ):

        self.meta = {}
        self.datamodule = None
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader
        self.num_classes = num_classes

        if (datamodule or train_dataloader) is None:
            raise UserWarning("Both datamodule and train_dataloader can't be None!")

        if all((datamodule, train_dataloader)):
            logger.warning("Both datamodule and train_dataloader is set! Using datamodule over train_dataloader.")

        if not datamodule:
            self.train_dataloader = train_dataloader
            self.val_dataloader = val_dataloader
            self.num_classes = num_classes

        elif isinstance(datamodule, pl.LightningDataModule):
            self.datamodule = datamodule
            self.train_dataloader = datamodule.train_dataloader()
            self.val_dataloader = datamodule.val_dataloader()
            if hasattr(datamodule, "num_classes"):
                self.num_classes = datamodule.num_classes
            if hasattr(datamodule, "num_labels"):
                self.meta["num_labels"] = datamodule.num_labels

        self.meta["num_classes"] = self.num_classes

    @property
    def sent_to_device(self):
        return self.meta.get("sent_to_device")

    @sent_to_device.setter
    def sent_to_device(self, value: bool = True):
        self.meta["sent_to_device"] = value

    def fetch(self, data, device_mapper: Optional[Callable] = None):
        """
        If data is not sent to `device` then will attempt to map the `device_mapper` function on data.
        Args:
            data: Data Batch
            device_mapper: Function to move data to device
        """
        if self.sent_to_device:
            return data
        if device_mapper:
            data = map(device_mapper, data)
        return data

    def get_train_dl(self, mapper_fn: Optional[Callable]):
        return self.fetch(self.train_dataloader, mapper_fn)

    def get_val_dl(self, mapper_fn: Optional[Callable]):
        return self.fetch(self.val_dataloader, mapper_fn)



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import math
from enum import Enum
from typing import Callable, Dict, Optional

import pytorch_lightning as pl
import torch

from gradsflow.callbacks import report_checkpoint_callback
from gradsflow.core.data import AutoDataset
from gradsflow.utility.common import module_to_cls_index

logger = logging.getLogger("core.backend")


class Backend(Enum):
    # Remove torch
    pl = "pl"
    gf = "gf"
    torch = "gf"
    default = "pl"


class AutoBackend:
    _OPTIMIZER_INDEX = module_to_cls_index(torch.optim, True)

    def __init__(
        self,
        autodataset: AutoDataset,
        model_builder: Callable,
        optimization_metric: Optional[str],
        max_epochs: int = 10,
        max_steps: Optional[int] = None,
        backend: Optional[str] = None,
    ):
        self.model_builder = model_builder
        self.backend = (backend or Backend.default.value).lower()
        self.autodataset = autodataset
        self.optimization_metric = optimization_metric
        self.max_epochs = max_epochs
        self.max_steps = max_steps

    def _gf_objective(
        self,
        search_space: Dict,
        trainer_config: Dict,
        gpu: Optional[float] = 0,
    ):

        autodataset = self.autodataset
        model = self.model_builder(search_space)
        tracker = model.fit(
            autodataset=autodataset,
            steps_per_epoch=self.max_steps,
            callbacks=trainer_config.get("callback_runner", ("tune_checkpoint", "tune_report")),
            show_progress=False,
            **trainer_config,
        )
        return tracker

    # noinspection PyTypeChecker
    def _lightning_objective(
        self,
        config: Dict,
        trainer_config: Dict,
        gpu: Optional[float] = 0,
    ):
        val_check_interval = 1.0
        if self.max_steps:
            val_check_interval = max(self.max_steps - 1, 1.0)

        datamodule = self.autodataset.datamodule

        trainer = pl.Trainer(
            logger=True,
            checkpoint_callback=False,
            gpus=math.ceil(gpu),
            max_epochs=self.max_epochs,
            max_steps=self.max_steps,
            callbacks=[report_checkpoint_callback()],
            val_check_interval=val_check_interval,
            **trainer_config,
        )

        model = self.model_builder(config)
        hparams = dict(model=model.hparams)
        trainer.logger.log_hyperparams(hparams)
        trainer.fit(model, datamodule=datamodule)

        logger.debug(trainer.callback_metrics)
        return trainer.callback_metrics[self.optimization_metric].item()

    def optimization_objective(self, config: dict, trainer_config: dict, gpu: Optional[float] = 0.0):
        """
        Defines lightning_objective function which is used by tuner to minimize/maximize the metric.

        Args:
            config dict: key value pair of hyperparameters.
            trainer_config dict: configurations passed directly to Lightning Trainer.
            gpu Optional[float]: GPU per trial
        """
        if self.backend == Backend.pl.value:
            return self._lightning_objective(config, trainer_config, gpu)

        if self.backend in (Backend.gf.value,):
            return self._gf_objective(config, trainer_config, gpu)

        raise NotImplementedError(f"Trainer not implemented for backend: {self.backend}")



"""An open-source AutoML Library based on PyTorch"""

#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from os import environ as _environ

_environ["LOGURU_LEVEL"] = _environ.get("LOGURU_LEVEL") or _environ.get("LOG_LEVEL", "ERROR")

from gradsflow.core.automodel import AutoModel
from gradsflow.core.data import AutoDataset
from gradsflow.models.model import Model
from gradsflow.tasks.autoclassification.image import AutoImageClassifier
from gradsflow.tasks.autoclassification.text import AutoTextClassifier
from gradsflow.tasks.autosummarization import AutoSummarization
from gradsflow.tasks.autotasks import autotask, available_tasks
from gradsflow.tuner.automodel import AutoModelV2
from gradsflow.tuner.tuner import Tuner

__version__ = "0.0.6.post1"



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import typing
from typing import Callable

if typing.TYPE_CHECKING:
    from gradsflow.models.model import Model


def dummy(x=None, *_, **__):
    return x


class Callback:
    """Callback objects define events on which it will run during the model training cycle."""

    _events = ("forward", "step", "epoch", "fit")

    def __init__(self, model: "Model"):
        self.model = model

    def with_event(self, event_type: str, func: Callable, exception, final_fn: Callable = dummy):
        """Calls a function with event wrapped around. Inspired from FastAI.
        Ref: https://github.com/fastai/fastai/blob/6e44b354f4d12bdfa2c9530f38f851c54a05764d/fastai/learner.py#L162
        """
        assert event_type in self._events, f"event_type is {event_type} but should be {self._events}"
        start_event = f"on_{event_type}_start"
        end_event = f"on_{event_type}_end"
        cancel_event = f"on_{event_type}_cancel"
        try:
            getattr(self, start_event)()
            func()
        except exception:
            getattr(self, cancel_event)()
        getattr(self, end_event)()
        final_fn()

    def on_fit_start(self):
        """Called on each `model.fit(...)`"""

    def on_fit_end(
        self,
    ):
        """Called after `model.fit(...)`"""

    def on_fit_cancel(
        self,
    ):
        """Called after `model.fit(...)`is cancelled"""

    def on_train_epoch_start(
        self,
    ):
        """Called on start of training epoch"""

    def on_train_epoch_end(self):
        """Called after end of training epoch"""

    def on_train_epoch_cancel(self):
        """Called after training epoch is cancelled"""

    def on_val_epoch_start(
        self,
    ):
        """Called on start of validation epoch"""

    def on_val_epoch_end(self):
        """called after validation epoch ends"""

    def on_val_epoch_cancel(self):
        """called after validation epoch cancelled"""

    def on_train_step_start(self):
        """called before `train_step`"""

    def on_train_step_end(self):
        """Called after training step"""

    def on_train_step_cancel(self):
        """Called after training step is cancelled"""

    def on_val_step_start(self):
        """Called on validation step"""

    def on_val_step_end(self):
        """Called after validation step"""

    def on_val_step_cancel(self):
        """Called after validation step is cancelled"""

    def on_epoch_start(self):
        """Called Before each Epoch"""

    def on_epoch_end(self):
        """Called after each epoch"""

    def on_epoch_cancel(self):
        """Called after epoch is cancelled"""

    def on_forward_start(self):
        """Called before model.forward(...)"""

    def on_forward_end(self):
        """Called after model.forward(...)"""

    def on_forward_cancel(self):
        """Called after model.forward(...) is cancelled"""

    def clean(self):
        """Clean up"""



"""Data loader for image dataset"""
#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
import os
from pathlib import Path
from typing import List, Optional, Tuple, Union

from torch.utils.data import DataLoader
from torchvision import transforms as T
from torchvision.datasets import ImageFolder

from gradsflow.core.data import Data
from gradsflow.data.ray_dataset import RayImageFolder

logger = logging.getLogger("data.image")


def get_augmentations(image_size: tuple = (224, 224), auto_augment_policy: bool = True):
    if auto_augment_policy:
        augmentations = [T.Resize(image_size), T.AutoAugment(), T.ToTensor()]
    else:
        augmentations = [T.Resize(image_size), T.ToTensor()]
    return T.Compose(augmentations)


def image_dataset_from_directory(
    directory: Union[List[str], Path, str],
    transform=None,
    image_size=(224, 224),
    batch_size: int = 1,
    shuffle: bool = False,
    pin_memory: bool = True,
    num_workers: Optional[int] = None,
    ray_data: bool = False,
) -> Data:
    """
    Create Dataset and Dataloader for image folder dataset.
    Args:
        directory:
        transform:
        image_size:
        batch_size:
        shuffle:
        pin_memory:
        num_workers:

    Returns:
        A dictionary containing dataset and dataloader.
    """
    data = Data()
    num_workers = num_workers or os.cpu_count()
    if transform is True:
        transform = get_augmentations(image_size)
    if ray_data:
        data.dataset = RayImageFolder(directory, transform=transform)
    else:
        data.dataset = ImageFolder(directory, transform=transform)
    logger.info("ds created")
    data.dataloader = DataLoader(
        data.dataset,
        batch_size=batch_size,
        pin_memory=pin_memory,
        shuffle=shuffle,
        num_workers=num_workers,
    )
    return data


def get_fake_data(
    image_size: Tuple[int, int], num_classes=10, batch_size=1, pin_memory=False, shuffle=True, num_workers=0
):
    from torchvision.datasets import FakeData

    data = Data()

    transform = get_augmentations(
        image_size=image_size,
    )
    data.dataset = FakeData(size=100, image_size=[3, *image_size], num_classes=num_classes, transform=transform)
    data.dataloader = DataLoader(
        data.dataset,
        batch_size=batch_size,
        pin_memory=pin_memory,
        shuffle=shuffle,
        num_workers=num_workers,
    )

    return data



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import timm

from gradsflow.core.autoclassifier import AutoClassifier
from gradsflow.core.backend import Backend
from gradsflow.models.model import Model


# noinspection PyTypeChecker
class AutoImageClassifier(AutoClassifier):
    """
    Automatically find Image Classification Model

    Args:
        datamodule Optional[DataModule]: PL Lightning DataModule with `num_classes` property.
        train_dataloader Optional[DataLoader]: torch dataloader
        val_dataloader Optional[DataLoader]: torch dataloader
        num_classes Optional[int]: number of classes
        max_epochs [int]: default=10.
        n_trials [int]: default=100.
        optimization_metric [Optional[str]]: defaults None
        suggested_backbones Union[List, str, None]: defaults None
        suggested_conf [Optional[dict] = None]: This sets Trial suggestions for optimizer,
            learning rate, and all the hyperparameters.
        timeout [int]: Hyperparameter search will stop after timeout.
        backend Optional[str]: Training loop code. Defaults to None.

    Examples:
        ```python
            from flash.core.data.utils import download_data
            from flash.image import ImageClassificationData

            from gradsflow import AutoImageClassifier

            # 1. Create the DataModule
            download_data("https://pl-flash-data.s3.amazonaws.com/hymenoptera_data.zip", "./data")

            datamodule = ImageClassificationData.from_folders(
                train_folder="data/hymenoptera_data/train/",
                val_folder="data/hymenoptera_data/val/",
            )

            model = AutoImageClassifier(datamodule,
                                        max_epochs=10,
                                        optimization_metric="val_accuracy",
                                        timeout=300)
            model.hp_tune()
        ```
    """

    _DEFAULT_BACKBONES = ["ssl_resnet18", "ssl_resnet50"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, backend=Backend.gf.value)

    def build_model(self, config: dict) -> Model:
        """Build ImageClassifier model from `ray.tune` hyperparameter configs
        or via _search_space dictionary arguments.

        Arguments:
            backbone [str]: Image classification backbone name - resnet18, resnet50,...
            (Check Lightning-Flash for full model list)

            optimizer [str]: PyTorch Optimizers. Check `AutoImageClassification._OPTIMIZER_INDEX`
            learning_rate [float]: Learning rate for the model.
        """
        backbone = config["backbone"]

        cnn = timm.create_model(backbone, pretrained=True, num_classes=self.num_classes)
        model = Model(cnn)
        model.compile(
            loss="crossentropyloss", optimizer=config["optimizer"], learning_rate=config["lr"], metrics="accuracy"
        )
        return model



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Dict, Optional

from rich.progress import BarColumn, Progress, RenderableColumn, TimeRemainingColumn
from rich.table import Column

from .callbacks import Callback


class ProgressCallback(Callback):
    def __init__(self, model, progress_kwargs: Optional[Dict] = None):
        super().__init__(model)
        progress_kwargs = progress_kwargs or {}
        tracker = self.model.tracker
        self.bar_column = BarColumn(table_column=Column(ratio=1))
        self.table_column = RenderableColumn(tracker.create_table(), table_column=Column(ratio=2))

        self.progress = Progress(
            "[progress.description]{task.description}",
            self.bar_column,
            "[progress.percentage]{task.percentage:>3.0f}%",
            TimeRemainingColumn(),
            self.table_column,
            **progress_kwargs,
        )
        tracker.progress = self.progress
        self.fit_prog = None
        self.train_prog_bar = None
        self.val_prog_bar = None

    def on_fit_start(self):
        self.progress.start()
        epochs = self.model.tracker.max_epochs
        completed = self.model.tracker.current_epoch
        self.fit_prog = self.progress.add_task("[red]Progress...", total=epochs, completed=completed)

    def on_fit_end(self):
        self.progress.stop()

    def on_epoch_end(self):
        self.progress.update(self.fit_prog, advance=1)

    def on_train_epoch_start(self):
        n = len(self.model.tracker.autodataset.train_dataloader)
        self.train_prog_bar = self.progress.add_task("[green]Learning...", total=n)

    def on_train_epoch_end(self):
        self.progress.remove_task(self.train_prog_bar)
        self.table_column.renderable = self.model.tracker.create_table()

    def on_train_step_end(self):
        self.progress.update(self.train_prog_bar, advance=1)
        self.table_column.renderable = self.model.tracker.create_table()

    def on_val_epoch_start(self):
        val_dl = self.model.tracker.autodataset.val_dataloader
        if not val_dl:
            return
        n = len(val_dl)
        self.val_prog_bar = self.progress.add_task("[green]Validating...", total=n)

    def on_val_epoch_end(self):
        val_dl = self.model.tracker.autodataset.val_dataloader
        if not val_dl:
            return
        self.table_column.renderable = self.model.tracker.create_table()
        self.progress.remove_task(self.val_prog_bar)

    def on_val_step_end(self):
        self.progress.update(self.val_prog_bar, advance=1)

    def clean(self):
        self.progress.stop()



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, Optional, Union

import torch

from gradsflow.utility.common import module_to_cls_index


class BaseAutoModel(ABC):
    """
    The main class for AutoML which consists everything required for tranining a model -
    data, model and trainer.
    """

    _OPTIMIZER_INDEX = module_to_cls_index(torch.optim, True)

    @abstractmethod
    def _create_search_space(self) -> Dict[str, str]:
        """creates search space"""
        raise NotImplementedError

    @abstractmethod
    def build_model(self, search_space: dict):
        """Build model from dictionary _search_space"""
        raise NotImplementedError


@dataclass(init=False)
class TrackingValues:
    loss: Optional[float] = None
    steps: Optional[int] = None
    step_loss: Optional[float] = None
    metrics: Union[None, Dict[str, float]] = None

    def __init__(self):
        self.metrics = {}


@dataclass(init=False)
class BaseTracker:
    max_epochs: int = 0
    current_epoch: int = 0  # current train current_epoch
    current_step: int = 0  # current current_step
    steps_per_epoch: Optional[int] = None
    train: TrackingValues = TrackingValues()
    val: TrackingValues = TrackingValues()



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import inspect
import os
import sys
from glob import glob
from pathlib import Path
from typing import Any, List

from smart_open import open as smart_open


def download(path):
    """Read any filesystem or cloud file"""
    with smart_open(path) as fr:
        return fr.read()


def get_file_extension(path: str) -> str:
    """Returns extension of the file"""
    return os.path.basename(path).split(".")[-1]


def get_files(folder: str):
    """Fetch every file from given folder recursively."""
    folder = str(Path(folder) / "**" / "*")
    return glob(folder, recursive=True)


def module_to_cls_index(module, lower_key: bool = True) -> dict:
    """Fetch classes from module and create a Dictionary with key as class name and value as Class"""
    class_members = inspect.getmembers(sys.modules[module.__name__], inspect.isclass)
    mapping = {}
    for k, v in class_members:
        if lower_key:
            k = k.lower()
        mapping[k] = v

    return mapping


def listify(item: Any) -> List:
    """Convert any scalar value into list."""
    if not item:
        return []
    if isinstance(item, list):
        return item
    if isinstance(item, (tuple, set)):
        return list(item)
    if isinstance(item, (int, float, str)):
        return [item]
    try:
        return list(item)
    except TypeError:
        return [item]



"""Mimics torch.data.Dataset for ray.data integration"""

#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import io
from typing import Callable, List, Union

import ray
from PIL import Image
from torch.utils.data import IterableDataset


class RayDataset(IterableDataset):
    def __init__(self, path: Union[List[str], str]):
        self.path = path
        self.ds = ray.data.read_binary_files(path, include_paths=True)

    def __iter__(self):
        return self.ds.iter_rows()

    def __len__(self):
        return len(self.input_files)

    def map_(self, func, *args, **kwargs) -> None:
        """Inplace Map for ray.data
        Time complexity: O(dataset size / parallelism)

        See https://docs.ray.io/en/latest/data/dataset.html#transforming-datasets"""
        self.ds = self.ds.map(func, *args, **kwargs)

    def map_batch_(self, func, batch_size: int = 2, **kwargs) -> None:
        """Inplace Map for ray.data
        Time complexity: O(dataset size / parallelism)
        See https://docs.ray.io/en/latest/data/dataset.html#transforming-datasets"""
        self.ds = self.ds.map_batches(func, batch_size=batch_size, **kwargs)

    @property
    def input_files(self):
        return self.ds.input_files()


class RayImageFolder(RayDataset):
    """Read image datasets
    ```
        root/dog/xxx.png
        root/dog/xxy.png
        root/dog/[...]/xxz.png

        root/cat/123.png
        root/cat/nsdf3.png
        root/cat/[...]/asd932_.png
    ```
    """

    def __init__(self, path, transform: Union[Callable, None] = None):
        super().__init__(path)
        self.transform = transform

    @staticmethod
    def file_to_class(files: Union[str, List]):
        file_list = []
        if isinstance(files, (tuple, list)):
            for file in files:
                file_list.append(file.split("/")[-2])
            return file_list
        return files.split("/")[-2]

    def find_classes(self) -> List[str]:
        files = self.input_files
        return sorted(list(set(map(self.file_to_class, files))))

    def __iter__(self):
        for data in self.ds.iter_rows():
            x = Image.open(io.BytesIO(data[1]))
            target = self.file_to_class(data[0])
            if self.transform:
                x = self.transform(x)
            yield x, target



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import abstractmethod
from typing import Dict, List, Optional, Union

import pytorch_lightning as pl
import torch
from ray import tune
from torch.utils.data import DataLoader

from gradsflow.core.automodel import AutoModel
from gradsflow.utility.common import listify


class AutoClassifier(AutoModel):
    """Implements `AutoModel` for classification tasks."""

    _DEFAULT_BACKBONES = []

    def __init__(
        self,
        datamodule: Optional[pl.LightningDataModule] = None,
        train_dataloader: Optional[DataLoader] = None,
        val_dataloader: Optional[DataLoader] = None,
        num_classes: Optional[int] = None,
        max_epochs: int = 10,
        max_steps: int = 10,
        n_trials: int = 100,
        optimization_metric: Optional[str] = None,
        suggested_backbones: Union[List, str, None] = None,
        suggested_conf: Optional[dict] = None,
        timeout: int = 600,
        prune: bool = True,
        backend: Optional[str] = None,
    ):
        super().__init__(
            datamodule=datamodule,
            train_dataloader=train_dataloader,
            val_dataloader=val_dataloader,
            num_classes=num_classes,
            max_epochs=max_epochs,
            max_steps=max_steps,
            optimization_metric=optimization_metric,
            n_trials=n_trials,
            suggested_conf=suggested_conf,
            timeout=timeout,
            prune=prune,
            backend=backend,
        )

        if isinstance(suggested_backbones, (str, list, tuple)):
            self.suggested_backbones = listify(suggested_backbones)
        elif suggested_backbones is None:
            self.suggested_backbones = self._DEFAULT_BACKBONES
        else:
            raise UserWarning("Invalid suggested_backbone type!")

        self.num_classes = num_classes

    def forward(self, x):
        if not self.model:
            raise UserWarning("model not initialized yet, run `hp_tune()` first.")
        return self.model(x)

    # noinspection PyTypeChecker
    def _create_search_space(self) -> Dict[str, str]:
        """Create hyperparameter config from `ray.tune`

        Returns:
             key-value pair of `ray.tune` _search_space
        """
        trial_backbone = tune.choice(self.suggested_backbones)
        trial_lr = tune.loguniform(*self.suggested_lr)
        trial_optimizer = tune.choice(self.suggested_optimizers)
        hparams = {
            "backbone": trial_backbone,
            "lr": trial_lr,
            "optimizer": trial_optimizer,
        }
        return hparams

    @abstractmethod
    def build_model(self, config: dict) -> torch.nn.Module:
        """Every Task implementing AutoClassifier has to implement a
        build model method that can build `torch.nn.Module` from dictionary config
        and return the model.
        """



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest
import ray
from ray.tune.sample import Domain

from gradsflow.tuner.tuner import ComplexObject, Tuner

complex_object = ComplexObject()


def test_append():
    complex_object.append("test_append")
    assert "test_append" in ray.get(complex_object.values)


def test_get_complex_object():
    assert complex_object.get_complex_object(0) == "test_append"


def test_to_choice():
    assert isinstance(complex_object.to_choice(), Domain)


def test_update_search_space():
    tuner = Tuner()
    tuner.update_search_space("test_update_search_space", complex_object)
    assert isinstance(tuner.get_complex_object("test_update_search_space", 0), str)

    with pytest.raises(UserWarning):
        tuner.update_search_space("hello", "world")


def test_union():
    tuner = Tuner()
    tuner1 = Tuner()
    tuner1.choice("dropout", 0.1, 0.2, 0.3)
    tuner2 = tuner.union(tuner1)
    assert tuner2.value.get("dropout") is not None


def test_merge():
    tuner1 = Tuner()
    tuner1.choice("dropout", 0.1, 0.2, 0.3)
    tuner2 = Tuner()
    tuner2.choice("layers", 1, 2, 3)
    tuner3 = Tuner.merge(tuner1, tuner2)
    assert "layers" in tuner3.value


def test_suggest_complex():
    tuner = Tuner()
    tuner.suggest_complex("test_complex", "val1", "val2")
    assert "test_complex" in tuner.value


def test_scalar():
    tuner = Tuner()
    tuner.scalar("a", "b")
    assert tuner.get("a") == "b"


def test_get():
    tuner = Tuner()
    tuner.choice("optimizer", "val1", "val2")
    assert isinstance(tuner.get("optimizer"), Domain)

    tuner.suggest_complex("complex_opt", "val1")
    assert tuner.get("complex_opt").get_complex_object(0) == "val1"

    with pytest.raises(KeyError):
        tuner.get("random_key")



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from pathlib import Path

from gradsflow.core.data import Data
from gradsflow.data.image import image_dataset_from_directory

data_dir = Path.cwd()


# todo: remote dataset test
def test_image_dataset_from_directory():
    folder = f"{data_dir}/data/test-data-cat-dog-v0/cat-dog/"
    res1 = image_dataset_from_directory(folder, transform=True, ray_data=True)
    assert isinstance(res1, Data)
    res2 = image_dataset_from_directory(folder, transform=True, ray_data=False)
    assert isinstance(res2, Data)



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import torch

from gradsflow.utility.common import (
    download,
    get_file_extension,
    get_files,
    listify,
    module_to_cls_index,
)


def test_create_module_index():
    assert isinstance(module_to_cls_index(torch.optim), dict)


def test_get_files():
    assert len(get_files("./")) != 0


def test_get_file_extension():
    assert get_file_extension("image.1.png") == "png"


def test_download():
    assert "gradsflow" in (download("README.md")).lower()


def test_listify():
    assert listify(None) == []
    assert listify(1) == [1]
    assert listify((1, 2)) == [1, 2]
    assert listify([1]) == [1]
    assert listify({"a": 1}) == ["a"]



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path

from flash.core.data.utils import download_data

cwd = str(Path.cwd())

download_data("https://pl-flash-data.s3.amazonaws.com/hymenoptera_data.zip", f"{cwd}/data")

download_data(
    "https://github.com/gradsflow/test-data/archive/refs/tags/cat-dog-v0.zip",
    f"{cwd}/data",
)



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import torch
from flash.text.classification import TextClassifier

from gradsflow.core.autoclassifier import AutoClassifier


# noinspection PyTypeChecker
class AutoTextClassifier(AutoClassifier):
    """
    Automatically find Text Classification Model

    Examples:
        ```python
            from gradsflow import AutoTextClassifier

            from flash.core.data.utils import download_data
            from flash.text import TextClassificationData

            download_data("https://pl-flash-data.s3.amazonaws.com/imdb.zip", "./data/")
            datamodule = TextClassificationData.from_csv(
                "review",
                "sentiment",
                train_file="data/imdb/train.csv",
                val_file="data/imdb/valid.csv",
            )

            model = AutoTextClassifier(datamodule,
                                       suggested_backbones=['sgugger/tiny-distilbert-classification'],
                                       max_epochs=10,
                                       optimization_metric="val_accuracy",
                                       timeout=300)
            model.hp_tune()
        ```

    Arguments:
        datamodule Optional[DataModule]: PL Lightning DataModule with `num_classes` property.
        train_dataloader Optional[DataLoader]: torch dataloader
        val_dataloader Optional[DataLoader]: torch dataloader
        num_classes Optional[int]: number of classes
        max_epochs [int]: default=10.
        n_trials [int]: default=100.
        optimization_metric [Optional[str]]: defaults None
        suggested_backbones Union[List, str, None]: defaults None
        suggested_conf [Optional[dict] = None]: This sets Trial suggestions for optimizer,
            learning rate, and all the hyperparameters.
        timeout [int]: Hyperparameter search will stop after timeout.
    """

    _DEFAULT_BACKBONES = [
        "distilbert-base-uncased-finetuned-sst-2-english",
        "sgugger/tiny-distilbert-classification",
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        meta = self.auto_dataset.meta
        self.num_classes = meta.get("num_labels") or meta.get("num_classes")
        print("num_classes = ", self.num_classes)

    def build_model(self, config: dict) -> torch.nn.Module:
        """Build TextClassifier model from `ray.tune` hyperparameter configs
        or via _search_space dictionary arguments

        Arguments:
            backbone [str]: Image classification backbone name - resnet18, resnet50,...
            (Check Lightning-Flash for full model list)

            optimizer [str]: PyTorch Optimizers. Check `AutoImageClassification._OPTIMIZER_INDEX`
            learning_rate [float]: Learning rate for the model.
        """

        backbone = config["backbone"]
        optimizer = config["optimizer"]
        learning_rate = config["lr"]

        return TextClassifier(
            self.num_classes,
            backbone=backbone,
            optimizer=self._OPTIMIZER_INDEX[optimizer],
            learning_rate=learning_rate,
        )



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['gradsflow',
 'gradsflow.callbacks',
 'gradsflow.core',
 'gradsflow.data',
 'gradsflow.models',
 'gradsflow.tasks',
 'gradsflow.tasks.autoclassification',
 'gradsflow.tuner',
 'gradsflow.utility']

package_data = \
{'': ['*']}

install_requires = \
['accelerate>=0.5.0',
 'smart_open>=5.1',
 'lightning-flash[image,text]==0.5.0',
 'ray[default,tune]>=1.6',
 'rich',
 'loguru>=0.5']

extras_require = \
{'test': ['pytest', 'coverage']}

setup(name='gradsflow',
      version='0.0.6.post1',
      description='An open-source AutoML Library based on PyTorch',
      author='Aniket Maurya',
      author_email='aniket@gradsflow.com',
      url='https://github.com/gradsflow/gradsflow',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7',
     )



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

os.environ["GF_CI"] = "true"

from unittest.mock import MagicMock

from gradsflow.tasks import AutoSummarization


def test_build_model():
    datamodule = MagicMock()
    model = AutoSummarization(
        datamodule,
        max_epochs=1,
        timeout=5,
        suggested_backbones="sshleifer/distilbart-cnn-12-6",
        n_trials=1,
    )
    model_confs = {
        "backbone": model._DEFAULT_BACKBONES[-1],
        "optimizer": "adam",
        "lr": 1e-3,
    }
    model.build_model(model_confs)



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Callable, Dict, List, Union

import numpy as np
import torch
import torchmetrics
from torch import nn
from torchmetrics import Metric

from gradsflow.utility.common import module_to_cls_index

_nn_classes = module_to_cls_index(nn)
_tm_classes = module_to_cls_index(torchmetrics, lower_key=False)

losses: Dict[str, Callable] = {k: v for k, v in _nn_classes.items() if "loss" in k}

metrics: Dict[str, Metric] = {k: v for k, v in _tm_classes.items() if 65 <= ord(k[0]) <= 90}
metrics = {k.lower(): v for k, v in metrics.items()}


def get_device():
    return "cuda" if torch.cuda.is_available() else "cpu"


def to_item(data: Union[torch.Tensor, Any]) -> Union[int, float, str, np.ndarray]:
    if torch.is_tensor(data):
        if data.requires_grad:
            data = data.detach()
        data = data.cpu().numpy()
    return data


def available_losses() -> List[str]:
    return list(losses.keys())


def available_metrics() -> List[str]:
    return list(metrics.keys())



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional, Union

import pytorch_lightning as pl
from torch.utils.data import DataLoader

from .autoclassification.image import AutoImageClassifier
from .autoclassification.text import AutoTextClassifier
from .autosummarization import AutoSummarization

SUPPORTED_TASKS = {
    "image-classification": AutoImageClassifier,
    "text-classification": AutoTextClassifier,
    "summarization": AutoSummarization,
}


def available_tasks() -> List[str]:
    """Get a list of all available tasks."""
    return list(SUPPORTED_TASKS.keys())


def autotask(
    datamodule: Optional[pl.LightningDataModule] = None,
    train_dataloader: Optional[DataLoader] = None,
    val_dataloader: Optional[DataLoader] = None,
    num_classes: Optional[int] = None,
    task: Optional[str] = None,
    data_type: Optional[str] = None,
    max_epochs: int = 10,
    max_steps: int = 10,
    n_trials: int = 100,
    optimization_metric: Optional[str] = None,
    suggested_backbones: Union[List, str, None] = None,
    suggested_conf: Optional[dict] = None,
    timeout: int = 600,
    prune: bool = True,
):
    """

    Args:
        datamodule Optional[DataModule]: PL Lightning DataModule with `num_classes` property.
        train_dataloader Optional[DataLoader]: torch dataloader
        val_dataloader Optional[DataLoader]: torch dataloader
        num_classes Optional[int]: number of classes
        task Optional[str]: type of task. Check available tasks `availalbe_tasks()
        data_type Optional[str]: default=None. type of data - image, text or infer.
        max_epochs [int]: default=10.
        n_trials [int]: default=100.
        optimization_metric [Optional[str]]: defaults None
        suggested_backbones Union[List, str, None]: defaults None
        suggested_conf [Optional[dict] = None]: This sets Trial suggestions for optimizer,
            learning rate, and all the hyperparameters.
        timeout [int]: Hyperparameter search will stop after timeout.

    Returns:
        Implementation of `AutoModel` for the task type.
    """
    if not (task or data_type):
        raise UserWarning("either task or data_type must be set!")

    if task not in SUPPORTED_TASKS:
        raise UserWarning(f"Unknown task {task}, available tasks are {list(SUPPORTED_TASKS.keys())}")

    targeted_task = SUPPORTED_TASKS[task]

    return targeted_task(
        datamodule=datamodule,
        train_dataloader=train_dataloader,
        val_dataloader=val_dataloader,
        num_classes=num_classes,
        max_epochs=max_epochs,
        max_steps=max_steps,
        n_trials=n_trials,
        optimization_metric=optimization_metric,
        suggested_backbones=suggested_backbones,
        suggested_conf=suggested_conf,
        timeout=timeout,
        prune=prune,
    )



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
import torch
from flash.image import ImageClassificationData

from gradsflow.core.automodel import AutoModel
from gradsflow.core.backend import Backend

cwd = Path.cwd()
datamodule = ImageClassificationData.from_folders(
    train_folder=f"{cwd}/data/hymenoptera_data/train/",
    val_folder=f"{cwd}/data/hymenoptera_data/val/",
)


@patch.multiple(AutoModel, __abstractmethods__=set())
def test_auto_model():
    assert AutoModel(datamodule)


@patch.multiple(AutoModel, __abstractmethods__=set())
def test_build_model():
    model = AutoModel(datamodule)
    with pytest.raises(NotImplementedError):
        model.build_model({"lr": 1})


@patch.multiple(AutoModel, __abstractmethods__=set())
def test_create_search_space():
    model = AutoModel(datamodule)
    with pytest.raises(NotImplementedError):
        model._create_search_space()


@patch.multiple(AutoModel, __abstractmethods__=set())
@patch("gradsflow.core.backend.pl")
def test_objective(mock_pl):
    optimization_metric = "val_accuracy"
    model = AutoModel(
        datamodule,
        optimization_metric=optimization_metric,
        backend=Backend.pl.value,
    )

    model.backend.model_builder = MagicMock()
    trainer = mock_pl.Trainer = MagicMock()
    trainer.callback_metrics = {optimization_metric: torch.as_tensor([1])}

    model.backend.optimization_objective({}, {})


@patch.multiple(AutoModel, __abstractmethods__=set())
@patch("gradsflow.core.automodel.tune.run")
def test_hp_tune(
    mock_tune,
):
    automodel = AutoModel(datamodule)
    automodel._create_search_space = MagicMock()
    automodel.build_model = MagicMock()
    automodel.hp_tune(gpu=0, cpu=2)

    mock_tune.assert_called()



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from unittest.mock import MagicMock, Mock, patch

import pytest

from gradsflow.core.backend import AutoBackend

trainer_config = {"show_progress": False}


@patch("gradsflow.core.backend.pl")
def test_optimization_objective(mock_pl: Mock):
    dm = MagicMock()
    model_builder = MagicMock()

    # backend is pl
    autotrainer = AutoBackend(dm, model_builder, optimization_metric="val_accuracy", backend="pl")
    autotrainer.optimization_objective({}, trainer_config)
    mock_pl.Trainer.assert_called()

    # wrong backend is passed
    with pytest.raises(NotImplementedError):
        autotrainer = AutoBackend(
            dm,
            model_builder,
            optimization_metric="val_accuracy",
            backend="error",
        )
        autotrainer.optimization_objective({}, trainer_config)



#  Copyright (c) 2021 GradsFlow. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import inspect
import os
from typing import Any, Callable, Dict, List, Optional, Union

import torch
from loguru import logger
from torch import nn
from torchmetrics import Metric

from gradsflow.callbacks import Callback, CallbackRunner
from gradsflow.callbacks.progress import ProgressCallback
from gradsflow.core.data import AutoDataset
from gradsflow.data.base import DataMixin
from gradsflow.models.base import BaseModel
from gradsflow.models.exceptions import EpochCancel, FitCancel
from gradsflow.models.tracker import Tracker
from gradsflow.utility.common import listify, module_to_cls_index

METRICS_TYPE = Union[str, Metric, List[Union[str, Metric]], None]


class Model(BaseModel, DataMixin):
    """
    Model provide training functionality with `model.fit(...)` inspired from Keras

    Examples:
    ```python
    model = Model(cnn)
    model.compile("crossentropyloss", "adam", learning_rate=1e-3, metrics="accuracy")
    model.fit(autodataset)
    ```

    Args:
        learner: Trainable model
        accelerator_config: HuggingFace Accelerator config
    """

    TEST = os.environ.get("GF_CI", "false").lower() == "true"
    _OPTIMIZER_INDEX = module_to_cls_index(torch.optim, True)

    def __init__(
        self,
        learner: Union[nn.Module, Any],
        device: Optional[str] = None,
        use_accelerate: bool = True,
        accelerator_config: dict = None,
    ):
        accelerator_config = accelerator_config or {}
        super().__init__(
            learner=learner,
            device=device,
            use_accelerate=use_accelerate,
            accelerator_config=accelerator_config,
        )
        self.callback_runner: Optional[CallbackRunner] = None

    def forward_once(self, x) -> torch.Tensor:
        self.callback_runner.on_forward_start()
        x = self.forward(x)
        self.callback_runner.on_forward_end()
        return x

    def compile(
        self,
        loss: Union[str, nn.modules.loss._Loss] = None,
        optimizer: Union[str, Callable] = None,
        learning_rate: float = 3e-4,
        metrics: METRICS_TYPE = None,
        loss_config: Optional[dict] = None,
        optimizer_config: Optional[dict] = None,
    ) -> None:
        """
        Examples:
            ```python
            model = Model(net)
            model.compile(loss="crossentropyloss", optimizer="adam", learning_rate=1e-3, metrics="accuracy")
            ```
        Args:
            loss: name of loss or torch Loss class object. See `available_losses()`
            optimizer: optimizer name or `torch.optim.Optimizer` Class
            learning_rate: defaults to 1e-3
            metrics: list of metrics to calculate. See `available_metrics()`
            loss_config: Dict config if any to pass to loss function
            optimizer_config: Dict config if any to pass to Optimizer

        """
        loss_config = loss_config or {}
        optimizer_config = optimizer_config or {}

        if optimizer:
            optimizer_fn = self._get_optimizer(optimizer)
            optimizer = optimizer_fn(self.learner.parameters(), lr=learning_rate, **optimizer_config)
            self.optimizer = self.prepare_optimizer(optimizer)
        if loss:
            self.loss = self._get_loss(loss)(**loss_config)
        self.add_metrics(*listify(metrics))
        self._compiled = True

    def calculate_metrics(self, inputs, target) -> Dict[str, torch.Tensor]:
        self.metrics.update(inputs, target)
        return self.metrics.compute()

    def train_step(self, batch: Union[List[torch.Tensor], Dict[Any, torch.Tensor]]) -> Dict[str, torch.Tensor]:
        inputs = self.fetch_inputs(batch)
        target = self.fetch_target(batch)

        self.optimizer.zero_grad()
        logits = self.forward_once(inputs)
        loss = self.loss(logits, target)
        self.backward(loss)
        self.optimizer.step()

        self.tracker.track("train/step_loss", loss, render=True)
        return {"loss": loss, "metrics": self.calculate_metrics(logits, target)}

    def val_step(self, batch: Union[List[torch.Tensor], Dict[Any, torch.Tensor]]) -> Dict[str, torch.Tensor]:
        inputs = self.fetch_inputs(batch)
        target = self.fetch_target(batch)

        logits = self.forward_once(inputs)
        loss = self.loss(logits, target)

        self.tracker.track("val/step_loss", loss, render=True)
        return {"loss": loss, "metrics": self.calculate_metrics(logits, target)}

    def train_one_epoch(self, train_dataloader):

        tracker = self.tracker
        running_train_loss = 0.0
        tracker.train.steps = 0
        steps_per_epoch = tracker.steps_per_epoch

        for step, batch in enumerate(train_dataloader):

            # ----- TRAIN STEP -----
            self.callback_runner.on_train_step_start()
            outputs = self.train_step(batch)
            self.callback_runner.on_train_step_end()

            # ----- METRIC UPDATES -----
            self.tracker.train.step_loss = outputs["loss"].item()
            self.tracker.track_metrics(outputs.get("metrics", {}), mode="train", render=True)

            running_train_loss += self.tracker.train.step_loss
            tracker.train.steps += 1
            tracker.current_step += 1
            if self.TEST:
                break
            if steps_per_epoch and step >= steps_per_epoch:
                break
        self.tracker.track_loss(running_train_loss / (tracker.train.steps + 1e-9), mode="train")

    def val_one_epoch(self, val_dataloader):
        tracker = self.tracker
        running_val_loss = 0.0
        tracker.val.steps = 0
        for _, batch in enumerate(val_dataloader):
            # ----- VAL STEP -----
            self.callback_runner.on_val_step_start()
            outputs = self.val_step(batch)
            self.callback_runner.on_val_step_end()

            # ----- METRIC UPDATES -----
            loss = outputs["loss"].item()
            self.tracker.val.step_loss = loss
            self.tracker.track_metrics(outputs.get("metrics", {}), mode="val", render=True)
            running_val_loss += self.tracker.val.step_loss
            tracker.val.steps += 1
            if self.TEST:
                break
        tracker.track_loss(running_val_loss / (tracker.val.steps + 1e-9), "val")

    def _train_epoch_with_event(self):
        self.callback_runner.on_train_epoch_start()
        train_dataloader = self.tracker.autodataset.get_train_dl(self.send_to_device)
        self.train()
        self.train_one_epoch(train_dataloader)
        self.callback_runner.on_train_epoch_end()
        self.metrics.reset()

    def _val_epoch_with_event(self):
        self.callback_runner.on_val_epoch_start()
        autodataset = self.tracker.autodataset
        if not autodataset.val_dataloader:
            return
        val_dataloader = self.tracker.autodataset.get_val_dl(self.send_to_device)
        self.eval()
        self.val_one_epoch(val_dataloader)
        self.callback_runner.on_val_epoch_end()
        self.metrics.reset()

    def epoch(self):
        current_epoch, max_epochs = self.tracker.current_epoch, self.tracker.max_epochs
        for epoch in range(current_epoch, max_epochs):
            self.tracker.current_epoch = epoch

            self._train_epoch_with_event()
            self._val_epoch_with_event()

            if self.TEST:
                break

    def _fit_with_event(self):
        self.callback_runner.on_fit_start()
        self.callback_runner.with_event("epoch", self.epoch, EpochCancel)
        self.callback_runner.on_fit_end()

    def fit(
        self,
        autodataset: AutoDataset,
        max_epochs: int = 1,
        steps_per_epoch: Optional[int] = None,
        callbacks: Union[List[str], Callback] = None,
        resume: bool = True,
        show_progress: bool = True,
        progress_kwargs=None,
    ) -> Tracker:
        """
        Similar to Keras model.fit(...) it trains the model for specified epochs and returns Tracker object

        Examples:
            ```python
            autodataset = AutoDataset(train_dataloader, val_dataloader)
            model = Model(cnn)
            model.compile("crossentropyloss", "adam", learning_rate=1e-3, metrics="accuracy")
            model.fit(autodataset)
            ```

        Args:
            autodataset: AutoDataset object encapsulate dataloader and datamodule
            max_epochs: number of epochs to train
            steps_per_epoch: Number of steps trained in a single current_epoch
            callbacks: Callback object or string
            resume: Resume training from the last current_epoch
            show_progress: Enable to show training progress
            progress_kwargs: Arguments for rich.progress

        Returns:
            Tracker object
        """
        if not resume:
            self.tracker.reset()
        self.assert_compiled()
        callback_list = listify(callbacks)
        if show_progress:
            callback_list.append(ProgressCallback(self, progress_kwargs))
        self.callback_runner = CallbackRunner(self, *callback_list)
        self.tracker.autodataset = self.prepare_data(autodataset)
        self.tracker.steps_per_epoch = steps_per_epoch
        self.tracker.max_epochs = max_epochs

        try:
            self.callback_runner.with_event("fit", self._fit_with_event, FitCancel)
        except KeyboardInterrupt:
            logger.error("Keyboard interruption detected")
        finally:
            self.callback_runner.clean()

        return self.tracker
